import os
import pathlib

def load_redirects(redirects_path) -> dict:
    with open(redirects_path) as f:
        lines = f.readlines()

    d = {}
    for line in lines:
        tokens = line.split(' ')
        assert len(tokens) == 2
        route = tokens[0].strip()
        target = tokens[1].strip()
        d[route] = target
    return d



def read(path):
    with open(path) as f:
        return f.read()

def gen_paths(cwd):
    contents_path = cwd / 'content'
    templates_path = cwd / 'templates'
    static_path = cwd / 'static'
    redirects_path = cwd / 'redirects.txt'
    frozen_path = cwd / 'routes-freeze.json'
    return (contents_path, templates_path, static_path, redirects_path, frozen_path)

def smart_capitalize(sentence):
    blacklist = ['and', 'the', 'of', 'to', 'by', 'in', 'a', 'for', 'as', 'but', 'it']
    tokens = sentence.split(' ')
    capitalized_tokens = []
    for i, token in enumerate(tokens):
        if i == 0 or (token not in blacklist):
            capitalized_tokens.append(token.capitalize())
        else:
            capitalized_tokens.append(token)
    return ' '.join(capitalized_tokens)



from flask import Flask, render_template, url_for, redirect
from flask.views import View
from gunicorn.app.base import BaseApplication
from . import parser
from . import redirects
import os

DEBUG=False

class MarkdownView(View):
    '''MarkdownView View

    This class should only be instantiated through cls.add_view(), as an argument for app.add_url_rule().
    Arguments:
    html     -- string containing markdown parsed to html to be inserted in template
    template -- relative route to template file
    title    -- title to be used for route
    '''
    def __init__(self, html, template, title):
        self.html = html
        self.template = template
        self.title = title
    def dispatch_request(self):
        '''Renders the specified template with markdown inserted and with specified title'''
        return render_template(self.template, md_parsed=self.html, title=self.title)

class RedirectView(View):
    '''RedirectView View

    This class shoudl only be instantiated through cls.add_view(), as an argument for app.add_url_rule().
    Arguments:
    target -- url to redirect to
    '''
    def __init__(self, target):
        self.target_url = target
    def dispatch_request(self):
        return redirect(self.target_url)

def create_server(root_directory, **kwargs):
    '''Returns a flask app object with routes pointing to compiled markdown files

    Arguments:
    root_directory     -- path to root directory of project, containing template and static directories

    Keyword Arguments:
    default_title      -- default title given to pages without specified title (default 'my-site')
    write              -- if true, parser saves route tree to '<root>/routes-freeze.json' to be
        read from later (default False)
    load_from_frozen   -- if true, loads route tree from '<root>/routes-freeze.json' instead of
        compiling tree again (default False)
    custom_view_routes -- dictionary of custom <route, (view, *arguments)> (default {})
    custom_view_route_override -- if true, routes specified in custom_view_routes will override (default
        False)
    '''
    template_folder = root_directory / 'templates'
    static_folder = root_directory / 'static'
    contents_folder = root_directory / 'contents'
    redirects_path = root_directory / 'redirects.txt'
    frozen_path = root_directory / 'routes-freeze.json'
    app = Flask(__name__, template_folder=template_folder, static_folder=static_folder)

    default_title = kwargs.get('default_title', 'my-site')
    write = kwargs.get('write', False)
    load_from_frozen = kwargs.get('load_from_frozen', False)
    if load_from_frozen == True:
        load_from_frozen = frozen_path

    custom_view_routes = kwargs.get('custom_view_routes', {}) # Dictionary of routes and tuple of view and args
    custom_view_route_override = kwargs.get('custom_view_route_override', False)

    def route_to_funcname(route):
        return route.replace('/', '_').replace('-', '_')

    if load_from_frozen:
        tree = parser.parse_tree_load(load_from_frozen)
    else:
        if write:
            tree = parser.parse_tree(contents_folder, write=True)
        else:
            tree = parser.parse_tree(contents_folder) # Outputs a list of routes and their paired html

    routes_added = []
    for route, html, title, template in tree:
        tree_routes.append(route)
        if route in custom_view_routes.keys() and custom_view_route_override: continue
        if DEBUG:
            print('Route %s (%s): %s' % (route, title, template))
        if title:
            app.add_url_rule(route, view_func=MarkdownView.as_view(route_to_funcname(route), html, template, title=title))
        else:
            app.add_url_rule(route, view_func=MarkdownView.as_view(route_to_funcname(route), html, template, title=default_title))
        routes_added.append(route)

    redirects_dict = redirects.load_redirects(redirects_path)
    for partial_route in redirects_dict.keys():
        route = os.path.join('/redirect/', partial_route)
        if route in custom_view_routes.keys() and custom_view_route_override: continue
        if DEBUG:
            print('Route %s: Redirect %s' % (route, redirects_dict[partial_route]))
        app.add_url_rule(route, view_func=RedirectView.as_view(route_to_funcname(route), redirects_dict[partial_route]))
        routes_added.append(route)

    for route in custom_view_routes.keys():
        view, *arguments = custom_view_routes[route]
        if not custom_view_route_override and route in routes_added:
            continue
        app.add_url_rule(route, view_func=view.as_view(*arguments))

    @app.errorhandler(404)
    def page_not_found(e):
        return render_template('404.html'), 404

    return app

# Gunicorn stuff

class StandaloneApplication(BaseApplication):
    def __init__(self, app, options=None):
        self.options = options or {}
        self.application = app
        super().__init__()

    def load_config(self):
        config = {key: value for key, value in self.options.items()
            if key in self.cfg.settings and value is not None}
        for key, value in config.items():
            self.cfg.set(key.lower(), value)

    def load(self):
        return self.application

def gunicornize(app, **options):
    return StandaloneApplication(app, options)



import argparse
from pathlib import Path
import shutil
import os
import multiprocessing
from . import helper
from . import parser
from . import server

# Global vars

CWD = Path.cwd()
CONTENT_PATH, TEMPLATES_PATH, STATIC_PATH, REDIRECTS_PATH, FROZEN_PATH = helper.gen_paths(CWD)
INTERNAL_TEMPLATES_PATH = Path(__file__).parent / 'resources' / 'templates'
INTERNAL_STATIC_PATH = Path(__file__).parent / 'resources' / 'static'
INTERNAL_CONTENT_PATH = Path(__file__).parent / 'resources' / 'content'

# Parser setup

argparser = argparse.ArgumentParser(
    prog='kobo'
)
argparser.add_argument('command', choices=['new', 'server', 'compile'])
argparser.add_argument('-c', '--compile', action='store_true', help='[server] Precompiles markdown files and saves their paths in `routes-freeze.json` (to be loaded via -L)')
argparser.add_argument('-g', '--gunicorn', action='store_true', help='[server] Runs the server using gunicorn')
argparser.add_argument('-p', '--port', type=int, help='[server] Specifies the port the server runs on')
argparser.add_argument('-L', '--load', action='store_true', help='[server] Loads from existing `routes-freeze.json` instead of compiling markdown files on server startup')
argparser.add_argument('--title', type=str, help='[server] Sets the default title of pages without an explicitly specified title')
argparser.add_argument('-s', '--single-file', dest='singlefile', action='store_true', help='[compile] Compiles single file and outputs to stdout or specified file (-i required)')
argparser.add_argument('-i', '--input', dest='in', type=str, help='[compile -s] Specifies input file for compile -s')
argparser.add_argument('-o', '--output', dest='out', type=str, help='[compile -s] Specifies output destination for compile -s')
argparser.add_argument('-v', '--verbose', action='store_true', help='[compile,server] Displays debug messages')


# Actually parse the args
args = argparser.parse_args()
if args.command == 'new':
    shutil.copytree(INTERNAL_CONTENT_PATH, CONTENT_PATH)
    shutil.copytree(INTERNAL_TEMPLATES_PATH, TEMPLATES_PATH)
    shutil.copytree(INTERNAL_STATIC_PATH, STATIC_PATH)
    REDIRECTS_PATH.touch(exist_ok=True)
    print('Created new kobo project in %s' % str(CWD))
    exit(0)

if args.command == 'server':
    kwargs = {'write': args.compile, 'load_from_frozen': args.load, 'default_title': args.title}
    server_app = server.create_server(CWD, **kwargs)
    port = args.port if args.port else 8000
    if not args.gunicorn:
        server_app.run('0.0.0.0', port=port)
    else:
        options = {
            'bind': '0.0.0.0:%s' % port,
            'workers': min(((multiprocessing.cpu_count() * 2) + 1), 4),
            'timeout': 120
        }
        gunicorn_app = server.gunicornize(server_app, **options)
        gunicorn_app.run()

    exit(0)

if args.command == 'compile':
    if args.singlefile:
        if not args.in:
            raise argparse.ArgumentError('flag `--input` must be specified')
        res = parser.parse_single(args.in)
        if not args.out:
            print(res)
        else:
            with open(args.out, 'w') as f:
                f.write(res)
        exit(0)
    parser.parse_tree_save(CONTENT_PATH, FROZEN_PATH, verbose=args.verbose)
    print('Saved routes to `%s`' % str(FROZEN_PATH))
    exit(0)
